from PyQt5.QtWidgets import QWidget, QVBoxLayout, QSizePolicy
from PyQt5.QtCore import Qt, QSize
from imagekit.Viewer.View.display_label import DisplayLabel

class DisplayView(QWidget):
    def __init__(self, vm):
        super().__init__()
        self.vm = vm
        self.vm.image_ready.connect(self.update_image)
        layout = QVBoxLayout(self)
        self.image_label = DisplayLabel(self.vm)
        self.image_label.display_changed.connect(self.vm.display_changed)
        self.image_label.setObjectName("image_label")
        self.label_size = QSize(self.image_label.width(), self.image_label.height())
        #self.image_label.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        #self.image_label.setAlignment(Qt.AlignCenter)
        #self.image_label.setStyleSheet("background-color: lightgray;")
        layout.addWidget(self.image_label)
        #layout.setContentsMargins(10, 10, 10, 10)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def resizeEvent(self, event):
        super().resizeEvent(event) # Call the base class implementation
        new_size = self.image_label.size()
        self.vm.label_size_changed.emit(new_size)

    def update_image(self):
        pixmap_image = self.vm.pixmap_image
        zoom = self.vm.zoom
        delta = self.vm.delta
        fit = self.vm.isFit
        self.image_label.transform_image(pixmap_image=pixmap_image, zoom_factor=zoom, delta=delta, fit= fit)







from PyQt5.QtCore import QObject, pyqtSignal
from imagekit.Histogram.histogram_model import HistogramModel
from imagekit.events import EventTypes,ImageLoadedEvent, ResetEvent
from aidkit.event.Aggregator import EventAggregator

class HistogramVM(QObject):
    vm_updated = pyqtSignal()
    def __init__(self, event_aggregator: EventAggregator):
        super().__init__()
        self.event_aggregator = event_aggregator
        self.event_aggregator.subscribe(EventTypes.IMAGE_LOADED, self.image_loaded_received)
        self.event_aggregator.subscribe(EventTypes.RESET, self.reset_received)
        self._model = None
        self.histogram = 0
        self.max_value = 0
        self.max_bin = 0

    def image_loaded_received(self, event: ImageLoadedEvent):
        self._model = HistogramModel(event.raw_image)
        self.histogram = self._model.histogram
        self.max_value = self._model.max_value
        self.max_bin = self._model.max_bin
        self.update_view()

    def reset_received(self, event:ResetEvent):
        pass

    def update_view(self):
        self.vm_updated.emit()


from typing import Tuple
from allytools.units import Length, LengthUnit


def size_in_inches(size_in: Tuple[Length, Length]) -> Tuple[float, float]:
    w_len, h_len = size_in
    w_in = w_len.to(LengthUnit.INCH)
    h_in = h_len.to(LengthUnit.INCH)
    return float(w_in), float(h_in)

def extent_numeric(extent: Tuple[Length, Length, Length, Length], unit: LengthUnit) -> Tuple[float, float, float, float]:
    xmin = extent[0].to(unit)
    xmax = extent[1].to(unit)
    ymin = extent[2].to(unit)
    ymax = extent[3].to(unit)
    return float(xmin), float(xmax), float(ymin), float(ymax)


from PyQt5 import QtWidgets
from imagekit.Viewer.View.view_port import ViewPort
from imagekit.Viewer.View.info import Info

class Control(QtWidgets.QWidget):
    def __init__(self,vm) :
        super().__init__()
        self.vm = vm
        self.vm.visibility.connect(self.visibility_changed)
        self.view_port = ViewPort(self.vm.view_port_vm)
        self.info = Info(self.vm.info_vm)
        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.view_port)
        self.layout.addWidget(self.info)
        self.setLayout(self.layout)

    def visibility_changed(self, vis: bool):
        self.setVisible(vis)
        self.setEnabled(vis)  # Enable or disable interactivity
        if vis:
            self.setStyleSheet("")  # Reset to default style when visible
        else:
            self.setStyleSheet("background: transparent; border: none;")









from PyQt5.QtCore import QPoint

class ViewPort:
    def __init__(self, zoom: float, delta:QPoint, fit_to_screen:bool, centered:bool, reset:bool):
        self.zoom = zoom
        self.delta =delta

        self.fit_to_screen = fit_to_screen
        self.centered = centered
        self.reset = reset











import numpy as np
from typing import Tuple, Optional, Union

Number = Union[int, float]

def resolve_limits_from_params(
    values: np.ndarray,
    pmin: Optional[Union[Number, str]],
    pmax: Optional[Union[Number, str]],
    *,
    fallback: Tuple[float, float] = (0.0, 1.0),
) -> Tuple[float, float]:
    """
    Resolve vmin/vmax given numeric or percentile string specs.
    - pmin/pmax can be numbers or strings like 'p1', 'p99.5'.
    - Percentiles are computed from finite data only.
    - Returns a safe (vmin, vmax) with vmax > vmin.
    """
    finite = values[np.isfinite(values)]
    vmin: Optional[float] = float(pmin) if isinstance(pmin, (int, float)) else None
    vmax: Optional[float] = float(pmax) if isinstance(pmax, (int, float)) else None

    # Collect percentile queries in one shot
    qs: list[float] = []
    idx: list[str] = []  # 'min' or 'max'

    if vmin is None and isinstance(pmin, str) and pmin.lower().startswith("p") and len(pmin) > 1:
        q = float(pmin[1:])
        q = min(100.0, max(0.0, q))
        qs.append(q); idx.append("min")

    if vmax is None and isinstance(pmax, str) and pmax.lower().startswith("p") and len(pmax) > 1:
        q = float(pmax[1:])
        q = min(100.0, max(0.0, q))
        qs.append(q); idx.append("max")

    if qs and finite.size:
        qs_vals = np.percentile(finite, qs)
        # If only one percentile requested, qs_vals is scalar
        if np.isscalar(qs_vals):
            qs_vals = [float(qs_vals)]
        for tag, val in zip(idx, qs_vals):
            if tag == "min":
                vmin = float(val)
            else:
                vmax = float(val)

    # Fallbacks if still None
    if finite.size:
        if vmin is None:
            vmin = float(np.min(finite))
        if vmax is None:
            vmax = float(np.max(finite))
    else:
        if vmin is None:
            vmin = fallback[0]
        if vmax is None:
            vmax = fallback[1]

    # Safety & ordering
    if not np.isfinite(vmin) or not np.isfinite(vmax):
        vmin, vmax = fallback
    if vmax <= vmin:
        # Nudge to a tiny positive span
        eps = max(1e-12, abs(vmin) * 1e-12)
        vmax = vmin + eps

    return vmin, vmax



from PyQt5.QtCore import QObject, pyqtSignal
from imagekit.Viewer.VM.view_port import ViewPortVM
from imagekit.Viewer.VM.info import InfoVM

class ControlVM(QObject):
    visibility = pyqtSignal(bool)
    def __init__(self, _display_vm):
        super().__init__()
        self.display_vm = _display_vm
        self.view_port_vm = ViewPortVM(self.display_vm)
        self.info_vm = InfoVM()

    def image_loaded(self):
        self.visibility.emit(True)



from PyQt5.QtCore import QObject, pyqtSignal
from imagekit.Viewer.Model.view_port import ViewPort

class ViewPortVM(QObject):
    change_to_viewport = pyqtSignal()

    def __init__(self, display_vm):
        super().__init__()
        self.display_vm = display_vm
        self.display_vm.notify_viewport.connect(self.changes_from_display)
        self._is_updating_from_view = False
        self.model =  ViewPort(zoom = 1, delta= (0, 0), fit_to_screen=True, centered=True, reset=True)

    def changes_from_display(self, zoom, delta):
        if not self._is_updating_from_view:
            self.model.delta = delta
            self.model.zoom = zoom
            self.display_vm.delta = delta
            self.display_vm.zoom = zoom
            self.change_to_viewport.emit()  # update zoom view value

    def get_zoom(self):
        return self.model.zoom

    def fit_clicked(self):
        self.display_vm.isFit = True
        self.delta_zero() # reset the offset
        self.display_vm.image_ready.emit()

    def set_offset(self, delta):
        self.display_vm.isFit = False
        self.model.delta = delta
        self.display_vm.image_ready.emit()

    def set_zoom(self, zoom):
        self.display_vm.isFit = False
        self._is_updating_from_view = True
        if self.model.zoom != zoom:
            self.model.zoom = zoom
        self.display_vm.zoom = zoom
        self._is_updating_from_view = False
        self.display_vm.image_ready.emit()

    def delta_zero(self):
        self.model.delta.setX(0)
        self.model.delta.setY(0)
        self.display_vm.delta.setX(0)
        self.display_vm.delta.setY(0)





from importlib.metadata import version, PackageNotFoundError

_DIST_NAME = __package__ or "imagera"

try:
    __version__ = version(_DIST_NAME)
except PackageNotFoundError:
    __version__ = "0.0.0"
__all__ = ["__version__"]


from .control import Control
from .display import DisplayView
from .display_label import DisplayLabel
from .fit import Fit
from .image_viewer import ImageViewer
from .info import Info
from .view_port import ViewPort
from .zoom import Zoom


__all__ = ["Control", "DisplayView", "DisplayLabel", "Fit", "ImageViewer", "Info", "ViewPort", "Zoom"]


from .control import ControlVM
from .display import DisplayVM
from .image_viewer import ImageViewerVM
from .info import InfoVM
from .view_port import ViewPortVM



__all__ = ["ControlVM", "DisplayVM", "ImageViewerVM", "InfoVM", "ViewPortVM"]


from PyQt5.QtCore import QObject
from imagekit.Viewer.VM.display import DisplayVM
from imagekit.Viewer.VM.control import ControlVM
from imagekit.events import EventTypes, ImageLoadedEvent, ResetEvent
from aidkit.event.Aggregator import EventAggregator


class ImageViewerVM(QObject):
    def __init__(self, event_aggregator: EventAggregator):
        super().__init__()
        self.event_aggregator = event_aggregator
        self.event_aggregator.subscribe(EventTypes.IMAGE_LOADED, self.image_loaded_received)
        self.event_aggregator.subscribe(EventTypes.RESET, self.reset_received)
        self.display_vm = DisplayVM()
        self.control_vm = ControlVM(self.display_vm)

    def image_loaded_received(self, event: ImageLoadedEvent):
        self.display_vm.new_image(event.raw_image)
        self.control_vm.image_loaded()

    def reset_received(self, event:ResetEvent):
        self.display_vm.reset_received()


import numpy as np
from argparse import ArgumentTypeError
import cv2
from PyQt5.QtCore import QSize
from PyQt5.QtGui import QImage
from imagera.image.image_aid import is_valid_image

#TODO replace to image bundle
class Image:
    def __init__(self, raw_image:np.array): #TODO ImageType as parameter
        if is_valid_image(raw_image):
            self._raw_image = raw_image
            #self.image_type = image_type
        else:
            raise ArgumentTypeError("numpy array is not an image")


    def get_np_array(self):
        return self._raw_image

    def shape(self):
        if self._raw_image is not None:
            return self._raw_image.shape
        return None

    def get_qsize(self):
        qsize = QSize(self.shape()[1], self.shape()[0])
        return  qsize

    def get_qimage(self) -> QImage:
        if self._raw_image.ndim == 2:  # Grayscale
            height, width = self._raw_image.shape
            bytes_per_line = width  # 1 byte per pixel
            qimage = QImage(self._raw_image.data, width, height, bytes_per_line, QImage.Format_Grayscale8)
        elif self._raw_image.ndim == 3:  # Color (BGR or RGB)
            height, width, channels = self._raw_image.shape
            bytes_per_line = 3 * width  # 3 bytes per pixel (RGB)
            if channels == 3:  # Assume BGR from OpenCV, convert to RGB
                rgb_image = cv2.cvtColor(self._raw_image, cv2.COLOR_BGR2RGB)
                qimage = QImage(rgb_image.data, width, height, bytes_per_line, QImage.Format_RGB888)
            else:
                raise ValueError("Unsupported channel size: Must be 1 or 3.")
        else:
            raise ValueError("Unsupported image format!")
        return qimage


from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

def fit_title(fig, text: str, *,
                  max_width_frac: float = 0.98,
                  max_height_frac: float = 0.08,
                  min_fontsize: int = 6,
                  max_fontsize: int = 24,
                  y_top: float = 0.995):
    """
    Create a suptitle that auto-shrinks to fit figure width/height and avoids clipping.
    Returns the Text instance (or None).
    """
    if not text:
        return None

    # Ensure we have a canvas/renderer
    canvas = FigureCanvas(fig)

    # Start large and shrink to fit
    t = fig.suptitle(text, y=y_top, fontsize=max_fontsize)
    canvas.draw()
    renderer = canvas.get_renderer()

    fig_w_px, fig_h_px = fig.get_size_inches() * fig.dpi
    fs = max_fontsize
    bb = t.get_window_extent(renderer=renderer)

    # shrink if too wide or too tall a band
    while fs > min_fontsize and (
        bb.width > max_width_frac * fig_w_px or
        bb.height > max_height_frac * fig_h_px
    ):
        fs -= 1
        t.set_fontsize(fs)
        canvas.draw()
        bb = t.get_window_extent(renderer=renderer)

    # If top is beyond the canvas, move it down a bit
    if bb.y1 > fig_h_px:
        overshoot = bb.y1 - fig_h_px
        new_y = 1.0 - (overshoot + 2.0) / fig_h_px  # 2 px padding
        # keep within a sensible band
        t.set_y(max(0.94, min(y_top, new_y)))
        canvas.draw()

    return t



import numpy as np

def is_valid_image(array: np.ndarray) -> bool:
    if array.ndim < 2: # Check if an array is at least 2D
        return False
    if array.dtype not in [np.uint8, np.float32, np.uint16]:
        return False
    if array.ndim == 2:  # Grayscale
        return True
    elif array.ndim == 3 and array.shape[2] in [3, 4]:  # RGB or RGBA
        return True
    return False


from dataclasses import dataclass
from typing import Optional, Tuple, Union, Any
from allytools.units import Length, LengthUnit
from imagera.plotter.cmap import CMaps
from matplotlib.colors import Colormap

Number = float
Percentile = str

@dataclass(frozen=True)
class PlotParameters:
    size_in: Tuple[Length, Length] = (
        Length(4.0, LengthUnit.INCH),
        Length(4.0, LengthUnit.INCH))
    dpi: int = 150
    use_levels: bool = True
    n_levels: int = 100
    cmap: Union[str, Colormap, CMaps] = CMaps.JET
    with_colorbar: bool = False
    hide_ticks: bool = True
    interpolation: Optional[str] = None
    x_label: Optional[str] = None
    y_label: Optional[str] = None
    value_label: Optional[str] = None
    plot_label: Optional[str] = None
    v_min: Optional[Union[Number, Percentile]] = None
    v_max: Optional[Union[Number, Percentile]] = None
    extent: Optional[Tuple[Length, Length, Length, Length]] = None
    show_lattice: bool = False
    lattice_pitch: Optional[Length] = None
    lattice_color: Any = (1, 1, 1, 0.4)

    def __post_init__(self) -> None:
        for size in self.size_in:
            if size.original_unit() is not LengthUnit.INCH:
                raise ValueError(f"PlotParameters.size_in must be in INCH units, "
                    f"got {size.original_unit()!r}")
        if self.extent is not None:
            units = {v.original_unit() for v in self.extent}

            if len(units) != 1:
                raise ValueError(f"All Length elements in extent must use the same unit, got {units!r}")
            (extent_unit,) = units




from typing import Tuple

def extent_from_affine(shape: Tuple[int, int], x0: float, y0: float, sx: float, sy: float):
    ny, nx = shape
    return x0, x0 + sx * nx, y0, y0 + sy * ny

def coerce_label(val):
    if val is None: return None
    s = str(val).strip()
    return None if s == "" or s.lower() == "none" else s

def is_percentile(x) -> bool:
    return isinstance(x, str) and x.lower().startswith("p") and len(x) > 1

def parse_percentile(x: str) -> float:
    return float(x[1:]) # "p1" -> 1.0, "p99.5" -> 99.5



from PyQt5.QtGui import QPixmap,  QPainter
from PyQt5.QtWidgets import QLabel
from PyQt5.QtCore import Qt, pyqtSignal, QPoint

class DisplayLabel(QLabel):
    display_changed = pyqtSignal(float, QPoint)
    def __init__(self, display_vm):
        super().__init__()
        self.display_vm = display_vm
        self.setMouseTracking(True)  # Enable mouse tracking without pressing any buttons
        self.start_pos = None  # Store the initial mouse press position
        self.zoom = 1.0
        self.delta = QPoint(0,0)
        self.offset_x = 0
        self.offset_y = 0
        self.mouse_delta = QPoint(0,0)
        self.pixmap = None

    def transform_image(self, pixmap_image: QPixmap, zoom_factor: float, delta:QPoint, fit:bool):
        self.pixmap = pixmap_image
        self.fit = fit
        if self.fit:
            self.zoom = self.fitted_zoom()
        else:
            self.zoom = zoom_factor
        self.delta = delta
        self.display_changed.emit(self.zoom, self.delta)
        self.update()  # Triggers paintEvent()

    def paintEvent(self, event):
        super().paintEvent(event)  # Call the base class paintEvent
        if self.pixmap:
            painter = QPainter(self)
            painter.setRenderHint(QPainter.SmoothPixmapTransform)  # Smooth scaling
            label_width = self.width()
            label_height = self.height()
            pixmap_width = self.pixmap.width()
            pixmap_height = self.pixmap.height()
            default_x = (label_width / self.zoom - pixmap_width)  / 2
            default_y = (label_height /self.zoom - pixmap_height) / 2
            offset_x = (default_x + self.delta.x() / self.zoom)
            offset_y = (default_y + self.delta.y() / self.zoom)
            painter.scale(self.zoom, self.zoom)
            painter.drawPixmap(int(offset_x), int(offset_y), self.pixmap)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and self.is_click_on_image(event.pos()):
            self.start_pos = event.pos()  - self.delta

    def mouseMoveEvent(self, event):
        pass
        if self.start_pos is not None:
            current_pos = event.pos()   #Get the current mouse position
            self.mouse_delta = current_pos - self.start_pos #Calculate movement in X and Y

            self.transform_image(pixmap_image = self.pixmap, zoom_factor = self.zoom, delta = self.mouse_delta, fit = False)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            if self.mouse_delta:
                self.delta = self.mouse_delta #Save achieved offset, only if it was some movement
            self.mouse_delta = QPoint(0,0)
            self.start_pos = None

    def wheelEvent(self, event):
        angle_delta = event.angleDelta()
        delta_y = angle_delta.y()
        factor = 0.1
        if delta_y > 0:
            self.zoom = self.zoom + factor
        elif delta_y < 0:
            self.zoom = self.zoom  - factor
        self.transform_image(pixmap_image=self.pixmap,
                             zoom_factor=self.zoom,
                             delta=QPoint(int(self.offset_x), int(self.offset_y)),
                             fit=False)

    def fitted_zoom(self) -> float:
        image_width = self.pixmap.width()
        image_height = self.pixmap.height()
        label_width = self.width()
        label_height = self.height()
        scale_w = label_width / image_width
        scale_h = label_height / image_height
        zoom_factor = min(scale_w, scale_h)
        return zoom_factor

    def is_click_on_image(self, click_pos: QPoint) -> bool:
        image_width = self.pixmap.width() * self.zoom
        image_height = self.pixmap.height() * self.zoom
        label_width, label_height = self.width(), self.height()
        image_x = (label_width - image_width) / 2 + self.delta.x()
        image_y = (label_height - image_height) / 2 + self.delta.y()
        return (
            image_x <= click_pos.x() <= image_x + image_width
            and image_y <= click_pos.y() <= image_y + image_height
        )


from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, Tuple
import numpy as np
import io
import base64

try:
    from PIL import Image
except ImportError:
    Image = None

# Optional deps – only used if available
try:
    import torch
except ImportError:
    torch = None

try:
    import tensorflow as tf
except ImportError:
    tf = None

@dataclass
class ImageBundle:
    """
    A thin adapter around a NumPy image that can produce
    multiple formats on demand (with simple caching).
    """
    # Accepts (H,W), (H,W,3) RGB, or (H,W,4) RGBA
    array: np.ndarray
    # If True, assume self.array in [0,1] float and auto-scale to uint8 for 8-bit formats
    assume_float_in_01: bool = False

    _cached_pil: Optional["Image.Image"] = field(default=None, init=False, repr=False)
    _cached_rgba8: Optional[np.ndarray] = field(default=None, init=False, repr=False)

    def __post_init__(self):
        a = self.array

        # ✅ Must be numpy
        if not isinstance(a, np.ndarray):
            raise TypeError(f"array must be a numpy.ndarray, got {type(a)}")

        # ✅ Shape checks
        if a.ndim == 2:
            # grayscale
            pass
        elif a.ndim == 3:
            if a.shape[2] not in (1, 3, 4):
                raise ValueError(
                    f"Third dimension must be 1, 3, or 4 channels, got {a.shape[2]}"
                )
        else:
            raise ValueError(
                f"array must be 2D (H,W) or 3D (H,W,C), got shape {a.shape}"
            )

        # ✅ dtype must be numeric (int or float)
        if not np.issubdtype(a.dtype, np.floating) and not np.issubdtype(a.dtype, np.integer):
            raise TypeError(
                f"Image array must have numeric dtype, got {a.dtype}"
            )

        # ✅ Optional float validation
        if np.issubdtype(a.dtype, np.floating) and self.assume_float_in_01:
            if np.nanmin(a) < 0.0 or np.nanmax(a) > 1.0:
                raise ValueError(
                    "assume_float_in_01=True but values are outside [0,1]"
                )

        # ✅ Channel=1 → treat as grayscale
        if a.ndim == 3 and a.shape[2] == 1:
            # collapse (H,W,1) → (H,W)
            self.array = a.reshape(a.shape[0], a.shape[1])


    def _to_rgba8(self) -> np.ndarray:
        """
        Returns an (H, W, 4) uint8 RGBA array.
        - Converts grayscale/RGB to RGBA.
        - Scales float arrays to uint8 if needed.
        """
        if self._cached_rgba8 is not None:
            return self._cached_rgba8

        a = self.array
        if a.ndim == 2:
            # Grayscale -> RGB
            a = np.stack([a, a, a], axis=-1)
        if a.ndim != 3 or a.shape[2] not in (3, 4):
            raise ValueError(f"Unsupported input shape: {self.array.shape}")

        # Handle dtype/scale
        if np.issubdtype(a.dtype, np.floating):
            # If values not in [0,1], try to auto-normalize per min/max
            if not self.assume_float_in_01:
                amin, amax = float(np.nanmin(a)), float(np.nanmax(a))
                if amax > amin:
                    a = (a - amin) / (amax - amin)
                else:
                    a = np.zeros_like(a)
            a = np.clip(a, 0.0, 1.0)
            a = (a * 255.0 + 0.5).astype(np.uint8)
        elif a.dtype != np.uint8:
            # Best-effort cast
            a = np.clip(a, 0, 255).astype(np.uint8)

        if a.shape[2] == 3:
            # Add opaque alpha
            alpha = np.full(a.shape[:2] + (1,), 255, dtype=np.uint8)
            a = np.concatenate([a, alpha], axis=2)

        self._cached_rgba8 = a
        return a

    def to_numpy(self, want: str = "rgba8") -> np.ndarray:
        """
        Return NumPy array in requested layout:
          - "rgb8"  -> (H,W,3) uint8
          - "rgba8" -> (H,W,4) uint8
          - "gray8" -> (H,W)   uint8 (luma from RGB)
        """
        rgba = self._to_rgba8()
        if want == "rgba8":
            return rgba
        if want == "rgb8":
            return rgba[..., :3]
        if want == "gray8":
            # simple luma approximation
            r, g, b = rgba[..., 0], rgba[..., 1], rgba[..., 2]
            y = (0.299 * r + 0.587 * g + 0.114 * b + 0.5).astype(np.uint8)
            return y
        raise ValueError(f"Unknown numpy format: {want}")

    def to_pillow(self) -> "Image.Image":
        """Return a Pillow Image (mode RGBA)."""
        if Image is None:
            raise RuntimeError("Pillow not installed.")
        if self._cached_pil is None:
            rgba = self._to_rgba8()
            self._cached_pil = Image.fromarray(rgba, mode="RGBA")
        return self._cached_pil

    def to_png_bytes(self, **pil_save_kwargs) -> bytes:
        """Return PNG bytes (lossless, supports alpha)."""
        im = self.to_pillow()
        buf = io.BytesIO()
        im.save(buf, format="PNG", **pil_save_kwargs)
        return buf.getvalue()

    def to_jpeg_bytes(self, quality: int = 90, **pil_save_kwargs) -> bytes:
        """Return JPEG bytes (lossy, no alpha)."""
        im = self.to_pillow().convert("RGB")
        buf = io.BytesIO()
        im.save(buf, format="JPEG", quality=int(quality), **pil_save_kwargs)
        return buf.getvalue()

    def to_tiff_bytes(self, **pil_save_kwargs) -> bytes:
        """Return TIFF bytes (can be lossless; good for high bit-depth)."""
        im = self.to_pillow()
        buf = io.BytesIO()
        im.save(buf, format="TIFF", **pil_save_kwargs)
        return buf.getvalue()

    def to_bmp_bytes(self, **pil_save_kwargs) -> bytes:
        """Return BMP bytes (un/low-compressed; legacy)."""
        im = self.to_pillow()
        buf = io.BytesIO()
        im.save(buf, format="BMP", **pil_save_kwargs)
        return buf.getvalue()

    def to_opencv_bgr(self) -> np.ndarray:
        """
        Return (H,W,3) uint8 in BGR order for OpenCV.
        Alpha channel (if any) is dropped.
        """
        rgb = self.to_numpy("rgb8")
        # RGB -> BGR
        return rgb[..., ::-1].copy()

    def to_torch_tensor(self, normalize_01: bool = True):
        """
        Return a torch.FloatTensor of shape (C,H,W).
        If normalize_01=True, scale to [0,1].
        """
        if torch is None:
            raise RuntimeError("PyTorch not installed.")
        rgb = self.to_numpy("rgb8").astype(np.float32)
        if normalize_01:
            rgb = rgb / 255.0
        t = torch.from_numpy(rgb)             # (H,W,C)
        return t.permute(2, 0, 1).contiguous()  # (C,H,W)

    def to_tf_tensor(self, normalize_01: bool = True):
        """
        Return a tf.float32 Tensor of shape (H,W,C).
        If normalize_01=True, scale to [0,1].
        """
        if tf is None:
            raise RuntimeError("TensorFlow not installed.")
        rgb = self.to_numpy("rgb8").astype(np.float32)
        if normalize_01:
            rgb = rgb / 255.0
        return tf.convert_to_tensor(rgb, dtype=tf.float32)

    def to_data_uri_png(self) -> str:
        """Return a data:image/png;base64,... string."""
        b = self.to_png_bytes()
        return "data:image/png;base64," + base64.b64encode(b).decode("ascii")

    def save(self, path: str, format: Optional[str] = None, **kwargs) -> None:
        """
        Save to disk using Pillow. Format inferred from extension if not given.
        """
        im = self.to_pillow()
        im.save(path, format=format, **kwargs)



from PyQt5.QtWidgets import QWidget, QVBoxLayout,QSizePolicy
from imagekit.Viewer.View.display import DisplayView
from imagekit.Viewer.View.control import Control

class ImageViewer(QWidget):
    def __init__(self, _vm, _control_size):
        super().__init__()
        self.vm = _vm
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.display = DisplayView(self.vm.display_vm)
        self.control = Control(self.vm.control_vm)
        self.display.setStyleSheet("border: 1px solid black;")
        self.display.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        layout.addWidget(self.display, stretch=1)
        layout.addWidget(self.control, stretch=0)
        #layout.addWidget(self.display)
        #   layout.addWidget(self.control)
        self.control.setFixedHeight(_control_size)
        self.control.setStyleSheet("background: transparent; border: none;")
        self.control.setDisabled(True)
        self.control.setVisible(True)

        self.setLayout(layout)









from __future__ import annotations
import math
from dataclasses import replace
from typing import Optional, Tuple
import numpy as np
import matplotlib.pyplot as plt
from allytools.units import LengthUnit
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.colorbar import Colorbar
from numpy.typing import NDArray
from imagera.plotter.aid import extent_numeric, size_in_inches
from imagera.plotter.plot_parameters import PlotParameters
from imagera.plotter.helpers import coerce_label
from imagera.plotter.cmap import resolve_cmap
from imagera.plotter.values_minmax import resolve_limits_from_params


class PltScalar2d:
    def __init__(
        self,
        data: NDArray[np.floating],
        params: Optional[PlotParameters] = None,
        **overrides,
    ):
        v = np.asarray(data, dtype=float)
        if v.ndim != 2:
            raise ValueError(f"'data' must be 2D, got shape {v.shape}")

        base = params if params is not None else PlotParameters()
        self.params: PlotParameters = replace(base, **overrides) if overrides else base
        self.data = v
        self.extent = self.params.extent
        self.v_min: Optional[float] = None
        self.v_max: Optional[float] = None
    def render(
        self,
        *,
        annotate_xy: Optional[tuple[float, float]] = None,
        plot_label: Optional[str] = None,
        **overrides,
    ) -> np.ndarray:
        """Standalone renderer: builds a Figure, draws, returns RGBA."""
        p = replace(self.params, **overrides)

        fig_size = size_in_inches(p.size_in)

        use_full_bleed = p.hide_ticks and not p.with_colorbar
        if use_full_bleed:
            fig = plt.figure(figsize=fig_size, dpi=p.dpi)
            ax = fig.add_axes((0, 0, 1, 1))
        else:
            fig, ax = plt.subplots(figsize=fig_size, dpi=p.dpi, layout="constrained")

        self.draw(
            ax,
            params=p,
            annotate_xy=annotate_xy,
            plot_label=plot_label,
            with_colorbar=p.with_colorbar)

        canvas = FigureCanvas(fig)
        canvas.draw()
        rgba = np.asarray(canvas.buffer_rgba()).copy()
        plt.close(fig)
        return rgba

    def render_into(
        self,
        ax,
        *,
        norm=None,
        annotate_xy=None,
        plot_label=None,
        with_colorbar=False,
        **overrides,
    ):
        """Inline renderer: draws into an existing Axes and returns AxesImage + Colorbar."""
        p = replace(self.params, **overrides)
        im, cbar = self.draw(
            ax,
            params=p,
            norm=norm,
            annotate_xy=annotate_xy,
            plot_label=plot_label,
            with_colorbar=with_colorbar)
        return im, cbar  # (AxesImage, Optional[Colorbar])

    def draw(
        self,
        ax,
        *,
        params: PlotParameters,
        norm=None,
        annotate_xy=None,
        plot_label=None,
        with_colorbar=False,
    ) -> Tuple[plt.AxesImage, Optional[Colorbar]]:
        """Draw the image and decorate the axes."""
        # Resolve extent: either user-provided (Length) or default pixel indices.
        unit = LengthUnit.UM
        if self.extent is None:
            ny, nx = self.data.shape
            extent = (0.0, float(nx), 0.0, float(ny))
            lattice_unit = None  # lattice pitch, if any, is interpreted directly as number of pixels
        else:
            extent = extent_numeric(self.extent, unit)
            lattice_unit = unit

        vmin, vmax = resolve_limits_from_params(self.data, params.v_min, params.v_max)

        cmap, norm = resolve_cmap(
            params.cmap,
            values=self.data,
            vmin=vmin,
            vmax=vmax,
            use_levels=params.use_levels,
            n_levels=params.n_levels)

        im = ax.imshow(
            self.data,
            extent=extent,
            origin="lower",
            aspect="equal",
            cmap=cmap,
            norm=norm,
            interpolation=params.interpolation)

        cbar: Optional[Colorbar] = None
        if with_colorbar:
            cbar = ax.figure.colorbar(im, ax=ax, pad=0.02)
            lbl = coerce_label(params.value_label)
            if lbl is not None:
                cbar.set_label(lbl)

        xlbl = coerce_label(params.x_label)
        ylbl = coerce_label(params.y_label)
        if xlbl:
            ax.set_xlabel(xlbl)
        if ylbl:
            ax.set_ylabel(ylbl)

        if params.hide_ticks:
            ax.set_xticks([])
            ax.set_yticks([])

        if annotate_xy is not None:
            ax.set_xlabel(
                f"({annotate_xy[0]:.3g}, {annotate_xy[1]:.3g})",
                labelpad=2,
                fontsize=9,
            )

        title = plot_label if plot_label is not None else params.plot_label
        if title:
            ax.set_title(str(title))

        if params.show_lattice and params.lattice_pitch is not None:
            pitch = float(params.lattice_pitch.to(unit))
            if pitch > 0.0:
                xmin, xmax, ymin, ymax = extent
                def first_tick(a, step):
                    return math.ceil(a / step) * step
                x0 = first_tick(xmin, pitch)
                y0 = first_tick(ymin, pitch)
                xs = np.arange(x0, xmax + 1e-12, pitch)
                ys = np.arange(y0, ymax + 1e-12, pitch)

                ax.vlines(
                    xs,
                    ymin,
                    ymax,
                    colors=params.lattice_color,
                    linewidth=0.5,
                    alpha=1.0,
                    zorder=5)
                ax.hlines(
                    ys,
                    xmin,
                    xmax,
                    colors=params.lattice_color,
                    linewidth=0.5,
                    alpha=1.0,
                    zorder=5)
        return im, cbar



from PyQt5 import QtWidgets

class Info(QtWidgets.QWidget):
    def __init__(self,vm) :
        super().__init__()
        self.vm = vm
        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)




import numpy as np
import  cv2

class HistogramModel:
    def __init__(self, _image:np.array):
        self.image = _image
        self.bit_depth = self._get_bit_depth()
        self.histogram = self._get_histogram()
        self.max_value = self.histogram.max()
        self.max_bin = np.argmax(self.histogram)

    def get_peak(self):
        pass

    def _get_bit_depth(self) -> int:
        return self.image.dtype.itemsize * 8

    def _get_histogram(self) -> np.ndarray:
        bins = 2 ** self.bit_depth
        range_max = 2 ** self.bit_depth
        return cv2.calcHist([self.image], [0], None, [bins], [0, range_max]).flatten() # TODO check channels



from PyQt5 import QtWidgets
from imagekit.Viewer.View.zoom import Zoom
from imagekit.Viewer.View.fit import Fit

class ViewPort(QtWidgets.QWidget):
    def __init__(self,vm) :
        super().__init__()
        self.vm = vm
        self.layout = QtWidgets.QHBoxLayout()
        self.zoom = Zoom(self.vm)
        self.fit = Fit(self.vm)
        self.layout.addWidget(self.zoom)
        self.layout.addWidget(self.fit)
        self.layout.addStretch(1)
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)

    def visibility_changed(self, vis: bool):
        self.setVisible(vis)










from __future__ import annotations
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from imagera.plotter import PltScalar2d
from imagera.plotter.aid import size_in_inches
from matplotlib.colors import Normalize, BoundaryNorm
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from dataclasses import replace
from typing import Optional, Tuple, Iterable, Union, Any
from imagera.plotter.plot_parameters import PlotParameters
from imagera.plotter.cmap import resolve_cmap
from imagera.plotter.fit_title import fit_title


def _as_xy_tuple(cell: Any) -> Tuple[float, float]:
    """
    Accept (x,y) in various forms:
      - tuple/list of length 2
      - ndarray shape (2,)
      - ndarray shape (1,2) or (2,1)
    Return (float(x), float(y)).
    """
    a = np.asarray(cell).reshape(-1)
    if a.size != 2:
        raise ValueError(f"Grid cell does not represent (x,y): got shape {np.asarray(cell).shape}")
    return float(a[0]), float(a[1])


def _iterate_tiles(values_array: np.ndarray) -> Iterable[Tuple[int, int, np.ndarray]]:
    """
    Yield (i, j, tile) where tile is a 2D array (Ny, Nx) for each field.
    Supports:
      - new numeric format: (R, C, Ny, Nx)
      - old object grid: (R, C) dtype=object with 2D arrays per cell
    """
    if values_array.ndim == 4:
        R, C, _, _ = values_array.shape
        for i in range(R):
            for j in range(C):
                yield i, j, np.asarray(values_array[i, j], dtype=float)
    elif values_array.ndim == 2 and values_array.dtype == object:
        R, C = values_array.shape
        for i in range(R):
            for j in range(C):
                tile = values_array[i, j]
                if tile is None:
                    yield i, j, None  # allow None (will render "No PSF")
                else:
                    yield i, j, np.asarray(tile, dtype=float)
    else:
        raise ValueError(
            "values_array must be either (R,C,Ny,Nx) numeric or (R,C) object grid of 2D arrays."
        )


def _deduce_field_shape(values_array: np.ndarray) -> Tuple[int, int]:
    if values_array.ndim == 4:
        return values_array.shape[:2]
    elif values_array.ndim == 2 and values_array.dtype == object:
        return values_array.shape
    else:
        raise ValueError("values_array must be either (R,C,Ny,Nx) numeric or (R,C) object grid.")


def plt_scalars2d(
    grid: np.ndarray,
    values_array: np.ndarray,
    *,
    params: Optional[PlotParameters] = None,
    mosaic_title: Optional[str] = None,
    **param_overrides
) -> np.ndarray:
    """
    Render an R×C mosaic of PSF tiles into a single RGBA uint8 image.
    Supports both the old object-grid format and the new stacked numeric format.
    """

    # 1️⃣ Determine R, C from values
    R, C = _deduce_field_shape(values_array)

    # Basic grid sanity: expect (R,C,2) numeric or (R,C) object/tuples that can become (x,y)
    if grid.shape[:2] != (R, C):
        raise ValueError(f"grid and values_array field shapes must match; got grid {grid.shape[:2]} vs values {(R, C)}")

    # 2️⃣ Extract function-only flags, build PlotParams
    tile_titles = bool(param_overrides.pop("tile_titles", False))
    with_colorbar = bool(param_overrides.pop("with_colorbar", True))
    mosaic_title = param_overrides.pop("mosaic_title", mosaic_title)

    base = params if params is not None else PlotParameters()
    p = replace(base, **param_overrides)
    cmap, _ = resolve_cmap(p.cmap)

    # 3️⃣ Global vmin/vmax across all tiles (ignore None)
    vmins, vmaxs = [], []
    for _, _, tile in _iterate_tiles(values_array):
        if tile is None:
            continue
        if np.isfinite(tile).any():
            vmins.append(np.nanmin(tile))
            vmaxs.append(np.nanmax(tile))
    if not vmins:
        raise ValueError("No valid PSF grids to render.")
    vmin = p.v_min if getattr(p, "v_min", None) is not None else float(min(vmins))
    vmax = p.v_max if getattr(p, "v_max", None) is not None else float(max(vmaxs))

    if p.use_levels:
        n_levels = max(2, int(p.n_levels))
        levels = np.linspace(vmin, vmax, n_levels)
        ncolors = getattr(cmap, "N", 256)
        norm = BoundaryNorm(levels, ncolors=ncolors, clip=True)
    else:
        norm = Normalize(vmin=vmin, vmax=vmax)

    # 4️⃣ Figure / GridSpec layout
    tile_w_in, tile_h_in =size_in_inches(p.size_in)
    extra_w = 0.6 if with_colorbar else 0.0
    fig_w_in = C * tile_w_in + extra_w
    fig_h_in = R * tile_h_in
    fig = plt.figure(figsize=(fig_w_in, fig_h_in), dpi=p.dpi)
    if with_colorbar:
        gs = gridspec.GridSpec(R, C + 1, figure=fig, width_ratios=[1] * C + [0.05])
        last_col_index = -1
    else:
        gs = gridspec.GridSpec(R, C, figure=fig, width_ratios=[1] * C)
        last_col_index = None

    # 5️⃣  Render (bottom-left = [0,0] in logical indices)
    mappable = None
    for i in range(R):
        # Logical row i=0 is bottom, but Matplotlib row 0 is top,
        # so map logical row i to plotting row (R-1-i).
        plot_row = R - 1 - i
        for j in range(C):
            ax = fig.add_subplot(gs[plot_row, j])

            # pick the tile for logical (i,j)
            if values_array.ndim == 4:
                tile = np.asarray(values_array[i, j], dtype=float)
            else:  # old object grid
                tile = values_array[i, j]
                tile = None if tile is None else np.asarray(tile, dtype=float)

            if tile is None:
                ax.text(0.5, 0.5, "No PSF", ha="center", va="center")
                ax.set_axis_off()
                continue
            dp = PltScalar2d(tile, params=p)

            # per-tile (x,y) annotation from logical grid (i,j)
            try:
                annotate_xy = (
                    None if getattr(p, "annotate_xy", None) is not None
                    else _as_xy_tuple(grid[i, j])
                )
            except Exception:
                annotate_xy = None

            plot_label = ("\u00A0" if not tile_titles else None)

            im, _ = dp.render_into(ax, norm=norm, annotate_xy=annotate_xy, plot_label=plot_label)
            if mappable is None:
                mappable = im

    # 6️⃣  Colorbar
    if with_colorbar and (mappable is not None) and (last_col_index is not None):
        cax = fig.add_subplot(gs[:, last_col_index])
        cbar = fig.colorbar(mappable, cax=cax)
        label = getattr(p, "value_label", None)
        if not label:
            # best-effort: look for a 'value_label' on first non-None tile-like object (old pipelines)
            if hasattr(values_array, "value_label"):
                label = getattr(values_array, "value_label")
        if label:
            cbar.set_label(str(label), fontsize=12)

    ## 7️⃣ Title & rasterize to RGBA
    if mosaic_title:
        # Leave some headroom for the title band
        fit_title(fig, str(mosaic_title))
        # Reserve space for suptitle so tight_layout doesn't collide with it
        fig.tight_layout(rect=(0.0, 0.0, 1.0, 0.95))
    else:
        fig.tight_layout()

    canvas = FigureCanvas(fig)
    canvas.draw()
    rgba = np.asarray(canvas.buffer_rgba()).copy()
    plt.close(fig)
    return rgba



from PyQt5.QtCore import QObject, pyqtSignal, QSize,QPoint
from PyQt5.QtGui import QPixmap
from imagekit.Viewer.Model.display import Display
import numpy as np

class DisplayVM(QObject):
    image_ready = pyqtSignal()
    label_size_changed = pyqtSignal(QSize)
    notify_viewport = pyqtSignal(float, QPoint)
    display_changed = pyqtSignal(float, QPoint)

    def __init__(self):
        super().__init__()
        self.display_changed.connect(lambda zoom, delta: self.notify_viewport.emit(zoom, delta))
        self.label_size_changed.connect(self.update_label_size)
        self.pixmap_image = None
        self.model = None

    def new_image(self, _image: np.array):
        self.model = Display(_image)  # TODO what happened with previous model?
        self.pixmap_image = QPixmap.fromImage(self.model.qimage)
        self.update_display()

    def reset(self):
        self.pixmap_image = QPixmap.fromImage(self.model.qimage)
        self.model.clean_annotate_image()
        self.update_display()

    def update_label_size(self, size: QSize):
        self.label_size = size

    def update_display(self):
        self.delta = QPoint(0, 0)  # TODO what happened with previous model?
        self.zoom = 1.0
        self.isFit = True
        self.image_ready.emit()


from allytools.event import EventAggregator, BasicEvents, EventTypes

import os, cv2
import numpy as np

class ImageReader:
    def __init__(self, event_aggregator: EventAggregator):
        self.event_aggregator = event_aggregator

    def read_image_from_file(self, file_path: str) -> np.ndarray:
        if not os.path.exists(file_path):
            self._publish_error(f"File '{file_path}' does not exist.")
        image = cv2.imread(file_path,cv2.IMREAD_UNCHANGED)
        if image is None:
            self._publish_error(f"Failed to load image from '{file_path}'. The file may not be a valid image.")
        return image

    def _publish_error(self, message: str) -> None:
        self.event_aggregator.publish(
            BasicEvents.ErrorOccurredEvent(
                event_type=EventTypes.ERROR_OCCURRED,
                message=message)
        )




from __future__ import annotations
from typing import Protocol, Any, Union
import numpy as np

class RendersRGBA(Protocol):
    def render(self, **overrides: Any) -> np.ndarray: ...

def _load_pillow():
    try:
        from PIL import Image  # type: ignore
    except Exception:
        raise RuntimeError("Pillow isn't installed. Run: pip install pillow")
    return Image

def to_pil_array(rgba: np.ndarray):
    """Convert an (H,W,4) uint8 array to a PIL Image."""
    Image = _load_pillow()
    if not (isinstance(rgba, np.ndarray) and rgba.ndim == 3 and rgba.shape[-1] == 4):
        raise ValueError("Expected an (H, W, 4) uint8 array.")
    if rgba.dtype != np.uint8:
        rgba = rgba.astype(np.uint8, copy=False)
    return Image.fromarray(rgba, mode="RGBA")

def render_to_pil(renderer: RendersRGBA, **overrides: Any):
    """Call renderer.render(**overrides) and convert to PIL Image."""
    arr = renderer.render(**overrides)
    return to_pil_array(arr)

#TODO change to Path
def save_png(obj: Union[np.ndarray, RendersRGBA], path: str, **overrides: Any):
    """
    Save either:
      - a NumPy RGBA array, or
      - a renderer with .render(**overrides) -> RGBA
    to a PNG file using PIL (lazy import).
    """
    if hasattr(obj, "render"):
        im = render_to_pil(obj, **overrides)
    else:
        im = to_pil_array(obj)  # type: ignore[arg-type]
    im.save(path)



from PyQt5.QtWidgets import QWidget, QVBoxLayout, QSizePolicy
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

class HistogramView(QWidget):
    def __init__(self, _vm):
        super().__init__()
        self.vm = _vm
        self.vm.vm_updated.connect(self.update_view)
        layout = QVBoxLayout()
        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)

        layout.addWidget(self.canvas)
        self.setLayout(layout)

    def update_view(self):
        self._draw_histogram()

    def _draw_histogram(self):
        self.figure.clear()
        ax = self.figure.add_subplot(111)

        # Plot histogram
        ax.plot(self.vm.histogram, color='blue')

        # Draw vertical line at peak
        peak_bin = self.vm.max_bin
        peak_value = self.vm.max_value
        ax.axvline(peak_bin, color='red', linestyle='--', linewidth=1)

        # Annotate peak
        annotation = f"Bin: {peak_bin}\nValue: {int(peak_value)}"
        ax.text(peak_bin + 10, peak_value * 0.8, annotation,
                fontsize=8, color='red', verticalalignment='top')

        # Axis labels
        ax.set_title("Image Histogram")
        ax.set_xlabel("Pixel Intensity")
        ax.set_ylabel("Frequency")

        self.canvas.draw()



from imagera.plotter.cmap import resolve_cmap, CMaps
from imagera.plotter.plot_parameters import PlotParameters
from imagera.plotter.plt_scalar2d import PltScalar2d
from imagera.plotter.plt_scalars2d import plt_scalars2d

__all__ = ["resolve_cmap", "CMaps", "PlotParameters", "PltScalar2d", "plt_scalars2d", ]


import numpy as np
from imagekit.Image.image import Image

class Display:
    def __init__(self, _image:np.array):
        #TODO add valid check from above
        self._root_image = Image(raw_image=_image)
        self.root_qsize = self._root_image.get_qsize()
        self.qimage = self._root_image.get_qimage()
        self.annotate_images = []

    def add_annotate_image(self, image:Image):
        self.annotate_images.append(image)

    def clean_annotate_image(self):
        self.annotate_images.clear()


from allytools.event import Event, EventType
from allytools.core import IterableConstants
import numpy as np
from dataclasses import dataclass

class EventTypes(IterableConstants):
    IMAGE_LOADED = EventType("IMAGE_LOADED", "Image successfully loaded -")
    RESET = EventType("RESET", "Reset to root image")

@dataclass
class ResetEvent(Event):
    pass

@dataclass
class ImageLoadedEvent(Event):
    file_path: str
    raw_image: np.array


from PyQt5 import QtWidgets, QtCore

def set_zoom(value):
    return int(value * 100.0)
def get_zoom(value):
    return value / 100.0

class Zoom(QtWidgets.QWidget):
    def __init__(self, viewport_vm):
        super().__init__()
        self.viewport_vm = viewport_vm
        self.viewport_vm.change_to_viewport.connect(self.update_value)
        self.layout = QtWidgets.QHBoxLayout()
        self.zoom_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.zoom_slider.setMinimum(10)
        self.zoom_slider.setMaximum(500)
        self.zoom_slider.setTickInterval(50)
        self.zoom_slider.setSingleStep(10)
        self.zoom_slider.setFixedWidth(250) #TODO
        self.zoom_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.zoom_slider.valueChanged.connect(self.on_slider_value_changed)
        self.label = QtWidgets.QLabel(f"{self.zoom_slider.value()}%")
        self.layout.addWidget(self.zoom_slider)
        self.layout.addWidget(self.label)
        self.layout.addStretch()
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)

    def update_value(self):
        new_value = set_zoom(self.viewport_vm.get_zoom())
        self.zoom_slider.blockSignals(True)
        self.zoom_slider.setValue(new_value)
        self.zoom_slider.blockSignals(False)
        self.label.setText(f"{self.zoom_slider.value()}%")

    def on_slider_value_changed(self, value):
        self.label.setText(f"{value}%")
        self.viewport_vm.set_zoom(get_zoom(value))







from PyQt5 import QtWidgets, QtCore
from aidkit.button.new_buton import new_button

class Fit(QtWidgets.QWidget):
    def __init__(self, viewport_vm):
        super().__init__()
        self.viewport_vm = viewport_vm
        self.layout = QtWidgets.QHBoxLayout()
        self.fit_button = new_button(self.layout, "F", self.button_clicked, True)
        self.fit_button.setFixedSize(50,50)
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)

    def button_clicked(self):
        self.viewport_vm.fit_clicked()







from __future__ import annotations
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Colormap, BoundaryNorm, Normalize
from typing import Optional, Union, Tuple
from enum import Enum

class CMaps(str, Enum):
    GRAY = "gray"
    INFERNO = "inferno"
    PLASMA = "plasma"
    VIRIDIS = "viridis"
    HOT = "hot"
    JET = "jet"
    CIVIDIS = "cividis"
    COOLWARM = "coolwarm"


def resolve_cmap(cmap: Union[str, Colormap, CMaps],
                 *,
                 values: Optional[np.ndarray] = None,
                 vmin: Optional[float] = None,
                 vmax: Optional[float] = None,
                 use_levels: bool = False,
                 n_levels: int = 100) -> Tuple[Colormap, Optional[Normalize]]:
    """
    Unified color and normalization resolver.

    Returns (cmap, norm)
    - cmap: matplotlib Colormap
    - norm: matplotlib Normalize or BoundaryNorm
    """
    # 1️⃣ Resolve the actual colormap
    if isinstance(cmap, CMaps):
        cmap = cmap.value
    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)

    # 2️⃣ Early return if no data (just the colormap)
    if values is None:
        return cmap, None

    # 3️⃣ Resolve limits
    finite = values[np.isfinite(values)]
    if vmin is None or vmax is None:
        if finite.size:
            if vmin is None: vmin = float(np.min(finite))
            if vmax is None: vmax = float(np.max(finite))
        else:
            vmin, vmax = 0.0, 1.0
    if not np.isfinite(vmin) or not np.isfinite(vmax) or vmax <= vmin:
        vmin, vmax = 0.0, 1.0

    # 4️⃣ Build normalization
    if use_levels:
        n_levels = max(2, int(n_levels))
        levels = np.linspace(vmin, vmax, n_levels)
        norm = BoundaryNorm(levels, ncolors=getattr(cmap, "N", 256), clip=True)
    else:
        norm = Normalize(vmin=vmin, vmax=vmax)

    return cmap, norm
